from unittest import TestCase
from django.test import RequestFactory
from grainy.core import (
    Namespace,
)
from grainy.const import (
    PERM_READ,
    PERM_DELETE,
    PERM_UPDATE,
    PERM_CREATE
)
from django_grainy.helpers import (
    namespace,
    int_flags,
    str_flags,
    dict_get_namespace,
    request_to_flag,
)

from django_grainy_test.models import(
    ModelA,
    ModelD
)

class TestHelpers(TestCase):

    def test_namespace(self):
        """
        test django_grainy.util.namespace
        """

        self.assertEqual(namespace(ModelA), "django_grainy_test.modela")
        self.assertEqual(namespace(ModelA()), "django_grainy_test.modela.none")
        self.assertEqual(namespace("a.b.c"), "a.b.c")
        self.assertEqual(namespace(None), "")
        with self.assertRaises(TypeError):
            namespace(object())

        self.assertEqual(namespace( (ModelA, "name") ), "django_grainy_test.modela.name")
        self.assertEqual(namespace( (ModelA(), "name") ), "django_grainy_test.modela.none.name")

        self.assertEqual(namespace(ModelD, value="parent"), "dynamic.parent")
        self.assertEqual(
            namespace(ModelD(), value="parent", other_value="child"),
            "dynamic.parent.child"
        )


    def test_request_to_flag(self):
        """
        test django_grainy.helpers.request_to_flag
        test django_grainy.helpers.request_method_to_flag
        """
        factory = RequestFactory()

        self.assertEqual(request_to_flag(factory.get("/view_class")), PERM_READ)
        self.assertEqual(request_to_flag(factory.put("/view_class")), PERM_UPDATE)
        self.assertEqual(request_to_flag(factory.post("/view_class")), PERM_CREATE)
        self.assertEqual(request_to_flag(factory.patch("/view_class")), PERM_UPDATE)
        self.assertEqual(request_to_flag(factory.delete("/view_class")), PERM_DELETE)


    def test_dict_get_namespace(self):
        namespace = Namespace("a.b.c")
        self.assertEqual(
            dict_get_namespace({"a":{"b":{"c":123}}}, namespace),
            123
        )
        with self.assertRaises(KeyError) as inst:
            self.assertEqual(
                dict_get_namespace({}, namespace),
                123
            )

    def test_int_flags(self):
        """
        test django_grainy.util.int_flags
        """

        self.assertEqual(int_flags("c"), PERM_CREATE)
        self.assertEqual(int_flags("cr"), PERM_CREATE | PERM_READ)
        self.assertEqual(int_flags("cru"), PERM_CREATE | PERM_READ | PERM_UPDATE)
        self.assertEqual(int_flags("crud"), PERM_CREATE | PERM_READ | PERM_UPDATE | PERM_DELETE)
        self.assertEqual(int_flags("xyz"), 0)
        self.assertEqual(int_flags(None), 0)

        self.assertEqual(int_flags(int_flags("c")), PERM_CREATE)
        self.assertEqual(int_flags(int_flags("cr")), PERM_CREATE | PERM_READ)
        self.assertEqual(int_flags(int_flags("cru")), PERM_CREATE | PERM_READ | PERM_UPDATE)
        self.assertEqual(int_flags(int_flags("crud")), PERM_CREATE | PERM_READ | PERM_UPDATE | PERM_DELETE)

        with self.assertRaises(TypeError):
            int_flags(object())

    def test_str_flags(self):
        """
        test django.grainy.util.str_flags
        """

        self.assertEqual(str_flags(PERM_READ), "r")
        self.assertEqual(str_flags(PERM_READ | PERM_UPDATE), "ru")
        self.assertEqual(str_flags(PERM_READ | PERM_CREATE), "cr")
        self.assertEqual(str_flags(PERM_READ | PERM_DELETE), "rd")






from grainy.core import (
    PermissionSet,
    Applicator
)

from django.db.models import QuerySet
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, AnonymousUser

from .helpers import (
    namespace,
    int_flags,
    str_flags
)
from .conf import (
    ANONYMOUS_PERMS
)

class Permissions(object):

    """
    A utility class that will allow you to perform permission checks
    for a user or group on cached permission sets
    """

    def __init__(self, obj):
        """
        Arguments:
            - obj <User|AnonymousUser|Group|Model>
        """

        if not hasattr(obj, "grainy_permissions") and not isinstance(obj, AnonymousUser):
            raise ValueError(
                "`obj` needs to be have a `grainy_permissions` relationship"
            )
        self.obj = obj
        self.pset = PermissionSet()
        self.applicator = Applicator(self.pset)
        self.loaded = False
        self.load()

        self.grant_all = (isinstance(obj, get_user_model()) and obj.is_superuser)

    def load(self, refresh=False):
        """
        Loads the permission set for the user or group specified in
        `self.obj`

        In case `self.obj` holds an `AnonymousUser` instance, perms are loaded
        from settings.GRAINY_ANONYMOUS_PERMS

        Keyword Arguments:
            - refresh <bool>: if True, permission set will be reloaded if it
                has been loaded before
        """
        if not hasattr(self.obj, "grainy_permissions"):
            if isinstance(self.obj, AnonymousUser):
                #Permission for AnonymousUser instance are loaded from
                #settigns
                if not self.loaded or refresh:
                    self.pset = PermissionSet(ANONYMOUS_PERMS)
                    self.loaded = True
            return

        if not self.loaded or refresh:
            self.pset = self.obj.grainy_permissions.permission_set()
            if isinstance(self.obj, get_user_model()):
                # if we are loading permissions for user, we need
                # to also merge in permissions from any of the
                # groups the user is a part of
                for group in self.obj.groups.all():
                    self.pset.update(
                        group.grainy_permissions.permission_set().permissions
                    )
            self.loaded = True

    def check(self, target, permissions, explicit=False, ignore_grant_all=False):
        """
        Check permissions for the specified target

        Arguments:
            - target <object|class|str>: check permissions to this object / namespace
            - permissions <int|str>: permission flags to check

        Keyword Arguments:
            - explicit <bool>: require explicit permissions to the complete target
                namespsace
            - ignore_grant_all <bool>: if True the `grant_all` property will be ignored
                during permission checks. If false, users with the `superuser` status will
                automatically pass all the permission checks.
        """
        if self.grant_all and not ignore_grant_all:
            return True
        return self.pset.check(namespace(target), int_flags(permissions), explicit=explicit)

    def get(self, target, as_string=False, explicit=False):
        """
        Returns the permission flags for the specified target

        Arguments:
            - target <object|class|str>: return permissions to this object /
                namespsace

        Keyword Arguments:
            - as_string <bool>: if True returns string flags instead of int flags
            - explicit <bool>: require explicit permissions to the complete target

        Returns:
            - <int>: permission flags
            - <str>: permission flags, if as_string=True
        """
        if as_string:
            return str_flags(self.pset.get_permissions(namespace(target), explicit=explicit))
        return self.pset.get_permissions(namespace(target), explicit=explicit)

    def apply(self, data):
        """
        Applies permissions to the specified data, removing all content that
        is not permissioned on a READ level

        Arguments:
            - data <dict>

        Returns:
            - dict: sanitized data
        """
        if self.grant_all:
            return data
        self.applicator.pset = self.pset
        return self.pset.apply(data, applicator=self.applicator)

    def instances(self, model, permissions, explicit=False, ignore_grant_all=False):
        """
        Return a list of all instances of the specified model that
        are permissioned at the specified level

        Arguments:
            - model <Model|QuerySet>
            - permissions <str|int>: permission flag(s)

        Keyword Arguments:
            - explicit <bool>: require explicit permissions to the complete target
                namespsace
            - ignore_grant_all <bool>: if True the `grant_all` property will be ignored
                during permission checks. If false, users with the `superuser` status will
                automatically pass all the permission checks.

        Returns:
            - list: model instances
        """

        if isinstance(model, QuerySet):
            q = model
        else:
            q = model.objects.all()

        return [
            instance for instance in q
            if self.check(
                instance,
                permissions,
                explicit=explicit,
                ignore_grant_all=ignore_grant_all
            )
        ]




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models

from django_grainy.decorators import grainy_model

from django_grainy.models import GrainyMixin, Permission, PermissionManager

# Create your models here.

"""
These are the models used during the django_grainy
unit tests. There is no need to ever install the "django_grainy_test"
app in your project
"""

class ModelBase(GrainyMixin, models.Model):
    class Meta(object):
        abstract = True

@grainy_model()
class ModelA(ModelBase):
    name = models.CharField(max_length=255)

@grainy_model(namespace="something.arbitrary")
class ModelB(ModelA):
    pass

@grainy_model(
    namespace=ModelB.Grainy.namespace(),
    namespace_instance=u"{namespace}.{instance.b.id}.c.{instance.id}"
)
class ModelC(ModelA):
    b = models.ForeignKey(ModelB, related_name="c", on_delete=models.CASCADE)

@grainy_model(
    namespace=u"dynamic.{value}",
    namespace_instance=u"{namespace}.{other_value}"
)
class ModelD(ModelA):
    pass


@grainy_model(namespace="x")
class ModelX(ModelA):
    pass

@grainy_model(namespace="custom", parent="x")
class ModelY(ModelA):
    x = models.ForeignKey(ModelX, related_name="y", on_delete=models.CASCADE)

@grainy_model(namespace="z", parent="y")
class ModelZ(ModelA):
    y = models.ForeignKey(ModelY, related_name="z", on_delete=models.CASCADE)


class APIKey(models.Model):
    key = models.CharField(max_length=255)

class APIKeyPermission(Permission):
    api_key = models.ForeignKey(APIKey, related_name="grainy_permissions", on_delete=models.CASCADE)
    objects = PermissionManager()




class DecoratorRequiresNamespace(ValueError):
    def __init__(self, decorator):
        super(DecoratorRequiresNamespace, self).__init__(
            "This decorator requires you to specify a namespace"
        )
        self.decorator = decorator



import json

from .util import UserTestCase
from django.test import Client, RequestFactory
from django.test.utils import override_settings
from django.db import models
from django.contrib.auth.models import AnonymousUser


from django_grainy.models import (
    UserPermission,
    GroupPermission,
    GrainyHandler,
    PermissionSet
)
from django_grainy.decorators import grainy_model

from django_grainy_test.models import (
    ModelA,
    ModelB,
    APIKey,
    APIKeyPermission
)

from django_grainy_test.views import (
    View,
    view,
    detail,
    Detail,
    JsonView
)

from django_grainy.util import (
    namespace,
    int_flags,
    str_flags,
    Permissions
)

from grainy.const import (
    PERM_READ,
    PERM_UPDATE,
    PERM_CREATE,
    PERM_DELETE,
)


class TestPermissions(UserTestCase):

    EXPECTED_PERMISSIONS_A = PermissionSet({
        ModelA.Grainy.namespace() : PERM_READ,
        ModelB.Grainy.namespace() : PERM_READ | PERM_UPDATE,
        view.Grainy.namespace() : PERM_READ,
        View.Grainy.namespace() : PERM_READ | PERM_CREATE | PERM_UPDATE | PERM_DELETE,
        "detail.1" : PERM_READ | PERM_CREATE | PERM_UPDATE | PERM_DELETE,
        "detail_manual.1" : PERM_READ | PERM_CREATE | PERM_UPDATE | PERM_DELETE,
        "detail_manual.get" : PERM_READ,
        "detail_manual" : PERM_CREATE,
    })

    GROUP_PERMISSIONS_A = PermissionSet({
        "secret.group" : PERM_READ
    })


    @classmethod
    def setUpTestData(cls):
        UserTestCase.setUpTestData()

        cls.api_key = APIKey.objects.create(key="test")

        cls.users["user_a"].grainy_permissions.add_permission_set(
            cls.EXPECTED_PERMISSIONS_A
        )

        cls.api_key.grainy_permissions.add_permission_set(
            cls.EXPECTED_PERMISSIONS_A
        )


        cls.users["user_b"].grainy_permissions.add_permission_set(
            {
                "x.y.z" : "r"
            }
        )

        cls.users["user_c"].grainy_permissions.add_permission_set(
            {
                "detail" : "r",
                "detail.1" : "crud",
                JsonView.Grainy.namespace(): "r",
                JsonView.Grainy.namespace("nested_dict.secret"): "r"
            }
        )

        cls.users["user_admin_a"].grainy_permissions.add_permission_set(
            {
                "detail.1" : "crud"
            }
        )

        cls.groups["group_a"].grainy_permissions.add_permission_set(
            cls.GROUP_PERMISSIONS_A
        )


    def test_permissions_init(self):
        """
        test django.grainy.util.Permissions.__init__
        """
        user = self.users["user_a"]
        group = self.groups["group_a"]
        api_key = self.api_key

        self.assertEqual(Permissions(user).obj, user)
        self.assertEqual(Permissions(group).obj, group)
        self.assertEqual(Permissions(api_key).obj, api_key)
        with self.assertRaises(ValueError):
            Permissions(object())

    def test_permissions_check(self):
        """
        test django.grainy.util.Permissions.check
        """
        user = self.users["user_a"]

        perms = Permissions(user)
        self.assertTrue(perms.check(ModelA, PERM_READ))
        self.assertFalse(perms.check(ModelA, PERM_UPDATE))
        self.assertTrue(perms.check(ModelB, PERM_READ))
        self.assertTrue(perms.check(ModelB, PERM_UPDATE))
        self.assertTrue(perms.check(ModelA(), PERM_READ))
        self.assertFalse(perms.check(ModelA(), PERM_UPDATE))
        self.assertTrue(perms.check(ModelB(), PERM_READ))
        self.assertTrue(perms.check(ModelB(), PERM_UPDATE))
        self.assertTrue(perms.check("secret.group", PERM_READ))
        self.assertFalse(perms.check("secret", PERM_READ))

        perms = Permissions(self.users["user_b"])
        self.assertTrue(perms.check("x.y.z", PERM_READ))
        self.assertFalse(perms.check("x.y.z", PERM_UPDATE))

        perms = Permissions(self.users["user_admin_a"])
        self.assertTrue(perms.check("x.y.z", PERM_READ))
        self.assertFalse(perms.check("x.y.z", PERM_READ, ignore_grant_all=True))

        perms = Permissions(self.api_key)
        self.assertTrue(perms.check(ModelA, PERM_READ))
        self.assertFalse(perms.check(ModelA, PERM_UPDATE))
        self.assertTrue(perms.check(ModelB, PERM_READ))
        self.assertTrue(perms.check(ModelB, PERM_UPDATE))
        self.assertTrue(perms.check(ModelA(), PERM_READ))
        self.assertFalse(perms.check(ModelA(), PERM_UPDATE))
        self.assertTrue(perms.check(ModelB(), PERM_READ))
        self.assertTrue(perms.check(ModelB(), PERM_UPDATE))



    def test_permissions_get(self):
        """
        test django.grainy.util.Permissions.get
        """

        perms = Permissions(self.users["user_a"])

        self.assertEqual(perms.get(ModelA), PERM_READ)
        self.assertEqual(perms.get(ModelA, as_string=True), "r")
        self.assertEqual(perms.get(ModelB), PERM_READ | PERM_UPDATE)
        self.assertEqual(perms.get(ModelB, as_string=True), "ru")
        self.assertEqual(perms.get("detail_manual", as_string=True), "c")
        self.assertEqual(perms.get("detail_manual.1", as_string=True, explicit=True), "crud")
        self.assertEqual(perms.get("detail_manual.2", as_string=True, explicit=True), "")
        self.assertEqual(perms.get("detail_manual.2", as_string=True), "c")



    def test_anonymous_permissions(self):
        user = AnonymousUser()
        perms = Permissions(user)
        self.assertTrue(perms.check("a.b.c", PERM_READ))
        self.assertTrue(perms.check("a.b.c.d", PERM_UPDATE | PERM_READ))
        self.assertFalse(perms.check("x.y.z", PERM_READ))


    def test_grainy_view(self):
        """
        test grainy_view decorator
        """

        ## test function view
        response = self.userclient("user_a").get("/view/")
        self.assertEqual(response.status_code, 200)

        response = self.userclient("user_b").get("/view/")
        self.assertEqual(response.status_code, 403)

        ## test class view
        for method in ["POST", "GET", "PUT", "PATCH", "DELETE"]:
            response = getattr(self.userclient("user_a"), method.lower())("/view_class/")
            self.assertEqual(response.status_code, 200)

            response = getattr(self.userclient("user_b"), method.lower())("/view_class/")
            self.assertEqual(response.status_code, 403)

        # test namespace formatting from request param
        client = self.userclient("user_a")

        response = client.get("/detail/1/")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content.decode("utf-8"), "ID 1")

        response = client.get("/detail/2/")
        self.assertEqual(response.status_code, 403)

        for method in ["get","post","delete","put","patch"]:
            response = getattr(client, method)("/detail_class/1/")
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.content.decode("utf-8"), "{} Response 1".format(method.upper()))

            response = getattr(client, method)("/detail_class/2/")
            self.assertEqual(response.status_code, 403)

        # test explicit view

        for username in ["user_c", "user_admin_a"]:

            response = self.userclient(username).get("/detail_explicit/1/")
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.content.decode("utf-8"), "ID 1")

            response = self.userclient(username).get("/detail_explicit/2/")
            self.assertEqual(response.status_code, 403)

            for method in ["get","post","delete","put","patch"]:
                response = getattr(self.userclient(username), method)("/detail_class_explicit/1/")
                self.assertEqual(response.status_code, 200)
                self.assertEqual(response.content.decode("utf-8"), "{} Response 1".format(method.upper()))

                response = getattr(self.userclient(username), method)("/detail_class_explicit/2/")
                self.assertEqual(response.status_code, 403)



        # test class manual view
        response = self.userclient("user_a").get("/detail_class_manual/1/")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content.decode("utf-8"), "GET Response 1")

        response = self.userclient("user_a").post("/detail_class_manual/")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content.decode("utf-8"), "POST Response 1")

        # test class view with request object in namespace
        # formatting
        response = self.userclient("user_a").get("/detail_class_reqfmt/1/")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content.decode("utf-8"), "GET Response 1")

        response = self.userclient("user_a").delete("/detail_class_reqfmt/1/")
        self.assertEqual(response.status_code, 403)


    def test_grainy_json_view(self):
        """
        test grainy_json_view decorator
        """

        factory = RequestFactory()

        request = factory.get("/view_class_json/")
        request.user = self.users["user_a"]
        response = JsonView.as_view()(request)
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            json.loads(response.content.decode("utf-8")),
            {"hello": "world", "nested_dict": {"public": "something"}}
        )

        request = factory.get("/view_class_json/")
        request.user = self.users["user_c"]
        response = JsonView().dispatch(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            json.loads(response.content.decode("utf-8")),
            {"hello": "world", "nested_dict": {"secret": "hidden", "public": "something"}}
        )

    def test_permissions_instances(self):
        """
        test util.Permissions.instances
        """

        perms_a = Permissions(self.users["user_a"])
        perms_b = Permissions(self.users["user_b"])
        perms_admin = Permissions(self.users["user_admin_a"])

        for i in range(1,4):
            ModelA.objects.create(name="Test {}".format(i))

        # user a should have read to all 3 instances of model a
        # through implicit permissions
        instances = perms_a.instances(ModelA, "r")
        self.assertEqual(len(instances), 3)

        # user as should have read to to 0 instances of model a
        # through explicit permissions
        instances = perms_a.instances(ModelA, "r", explicit=True)
        self.assertEqual(len(instances), 0)

        # add user a permission to first isntance of model a
        # and reload permissions util
        self.users["user_a"].grainy_permissions.add_permission(ModelA.objects.first(), "r")
        perms_a.load(refresh=True)

        # user a should now habe read to 1 instances of model a
        # through explicit permissions
        instances = perms_a.instances(ModelA, "r", explicit=True)
        self.assertEqual(len(instances), 1)

        # user b should not have any
        instances = perms_b.instances(ModelA, "r")
        self.assertEqual(len(instances), 0)

        # deny user a permissions to first instance of model a
        # and reload permissions util
        self.users["user_a"].grainy_permissions.add_permission(ModelA.objects.first(), 0)
        perms_a.load(refresh=True)

        # user a should now have read access to 2 instances of model a
        instances = perms_a.instances(ModelA, "r")
        self.assertEqual(len(instances), 2)

        # admin user should have read to all 3 instances of model a
        # through grant all
        instances = perms_admin.instances(ModelA, "r")
        self.assertEqual(len(instances), 3)

        # admin user should have read to 0 instances of model
        # when ignoring grant all
        instances = perms_admin.instances(ModelA, "r", ignore_grant_all=True)
        self.assertEqual(len(instances), 0)





# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib import admin
from django.contrib.auth.admin import UserAdmin, GroupAdmin
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group

from .conf import ADMIN_REMOVE_DEFAULT_FORMS
from .models import (
    UserPermission,
    GroupPermission
)
from .forms import (
    UserPermissionForm,
    GroupPermissionForm
)

# Register your models here.

class UserPermissionInlineAdmin(admin.TabularInline):
    model = UserPermission
    form = UserPermissionForm
    extra = 1
    ordering = ("namespace",)

class GroupPermissionInlineAdmin(admin.TabularInline):
    model = GroupPermission
    form = GroupPermissionForm
    extra = 1
    ordering = ("namespace",)

## INIT

admin.site.unregister(get_user_model())
admin.site.unregister(Group)

_fieldsets = UserAdmin.fieldsets
for name, info in _fieldsets:
    if ADMIN_REMOVE_DEFAULT_FORMS and "user_permissions" in info.get("fields",[]):
        lst = list(info.get("fields"))
        lst.remove("user_permissions")
        info["fields"] = lst

@admin.register(get_user_model())
class GrainyUserAdmin(UserAdmin):
    fieldsets = _fieldsets
    inlines = UserAdmin.inlines + [UserPermissionInlineAdmin]

_exclude = []
if ADMIN_REMOVE_DEFAULT_FORMS:
    _exclude.append("permissions")

@admin.register(Group)
class GrainyGroupAdmin(GroupAdmin):
    inlines = GroupAdmin.inlines + [GroupPermissionInlineAdmin]
    exclude = _exclude



import six
import inspect
from grainy.core import (
    Namespace,
)
from .conf import (
    PERM_CHOICES,
    REQUEST_METHOD_TO_FLAG,
    DJANGO_OP_TO_FLAG
)

def namespace(target, **kwargs):

    """
    Convert `target` to permissioning namespace

    Any keyword arguments will be used for formatting of the
    namespace (as applicable)

    Arguments:
        - target <object|class|string|tuple>: if an object or class is passed here it
            will be required to contain a `Grainy` meta class, otherwise a
            TypeError will be raised.

            If a tuple is passed all elements of the tuple will be passed to
            the namespace function individually and the resulting namespaces will
            be joined together and returned as one namespace string

    Returns:
        - string
    """

    if not target:
        return ""

    if isinstance(target, tuple) or isinstance(target, list):
        result = []
        for _t in target:
            result.append(namespace(_t))
        return str(Namespace(result))

    handler_class = getattr(target, "Grainy", None)

    if inspect.isclass(handler_class):
        if inspect.isclass(target):
            return target.Grainy.namespace(**kwargs)
        return target.Grainy.namespace(instance=target, **kwargs)


    if isinstance(target, six.string_types):
        return target

    raise TypeError("`target` {} could not be convered to a permissioning namespace".format(target))



def dict_get_namespace(data, namespace):
    d = data
    path = []
    for k in namespace:
        if k not in d:
            raise KeyError("`{}` does not exist at `{}`".format(
                k,
                ".".join(path)
            ))
        path.append(k)
        d = d[k]
    return d


def request_to_flag(request):
    """
    Returns the appropriate grainy permission flag for the request
    depending on the request's method.

    Arguments:
        - request <Request>: django request object

    Returns:
        - int
    """
    return request_method_to_flag(request.method)

def request_method_to_flag(method):
    """
    Converts a request method to the matching grainy permission
    flag

    Arguments:
        - method <str>: request method 'GET', 'POST' etc.

    Returns:
        - int
    """
    return REQUEST_METHOD_TO_FLAG.get(method.upper(), 0)

def django_op_to_flag(op):
    """
    Converts a django admin operation string to the matching
    grainy permission flag

    Arguments:
        - op <str>

    Returns:
        - int
    """
    return DJANGO_OP_TO_FLAG.get(op, 0)


def int_flags(flags):
    """
    Converts string permission flags into integer permission flags

    Arguments:
        - flags <str>: one or more flags as they are defined in GRAINY_PERM_CHOICES

            For example: "crud" or "ru" or "r"

    Returns:
        - int
    """

    r = 0
    if not flags:
        return r

    if isinstance(flags, six.integer_types):
        return flags

    if not isinstance(flags, six.string_types):
        raise TypeError("`flags` needs to be a string or integer type")

    for f in flags:
        for f_i, name, f_s in PERM_CHOICES:
            if f_s == f:
                r = r | f_i
    return r


def str_flags(flags):
    """
    Converts integer permission flags into string permission flags

    Arguments:
        - flags <int>: one or more flags as they are defined in GRAINY_PERM_CHOICES

            For example: PERM_READ or PERM_READ | PERM_UPDATE

    Returns:
        - str
    """

    r = ""
    if not flags:
        return r

    if isinstance(flags, six.string_types):
        return flags

    if not isinstance(flags, six.integer_types):
        raise TypeError("`flags` needs to be a string or integer type")

    for f_i, name, f_s in PERM_CHOICES:
        if flags & f_i:
            r += f_s
    return r







from rest_framework.test import APIClient
from django.test import RequestFactory
from .util import UserTestCase

from grainy.const import (
    PERM_READ,
    PERM_CREATE,
    PERM_UPDATE,
    PERM_DELETE
)

from grainy.core import (
    PermissionSet,
)

from django_grainy_test.views import (
    ModelAViewSet,
)

from django_grainy_test.models import (
    ModelA,
)

class TestGrainyViewSet(UserTestCase):

    @classmethod
    def setUpTestData(cls):
        UserTestCase.setUpTestData()

        cls.users["user_a"].grainy_permissions.add_permission_set(
            PermissionSet({
                "api" : PERM_READ | PERM_CREATE | PERM_UPDATE | PERM_DELETE,
                "api.a.*.nested_dict.secret": PERM_READ
            })
        )

        cls.users["user_b"].grainy_permissions.add_permission_set(
            PermissionSet({
                "api" : PERM_READ,
                "api.a_x" : PERM_READ
            })
        )

        ModelA.objects.create(name="Test1")


    def test_grainy_viewset_gate(self):

        client_a = APIClient()
        client_a.force_authenticate(user=self.users["user_a"])

        r = client_a.post("/a/", {"name":"bla"})
        self.assertEqual(r.status_code, 201)
        r = client_a.put("/a/2/", {"name":"bla2"})
        self.assertEqual(r.status_code, 200)
        r = client_a.patch("/a/2/", {"name":"bla3"})
        self.assertEqual(r.status_code, 200)
        r = client_a.delete("/a/2/", follow=True)
        self.assertEqual(r.status_code, 404)

        r = client_a.get("/a_x/1/")
        self.assertEqual(r.status_code, 403)
        r = client_a.delete("/a_x/1/")
        self.assertEqual(r.status_code, 403)


        client_b = APIClient()
        client_b.force_authenticate(user=self.users["user_b"])

        r = client_b.post("/a/", {"name":"bla"})
        self.assertEqual(r.status_code, 403)
        r = client_b.put("/a/1/", {"name":"bla2"})
        self.assertEqual(r.status_code, 403)
        r = client_b.patch("/a/1/", {"name":"bla3"})
        self.assertEqual(r.status_code, 403)
        r = client_b.delete("/a/1/", follow=True)
        self.assertEqual(r.status_code, 403)

        r = client_b.get("/a_x/1/")
        self.assertEqual(r.status_code, 200)
        r = client_b.delete("/a_x/1/")
        self.assertEqual(r.status_code, 403)





    def test_grainy_viewset_list(self):
        client = APIClient()
        client.force_authenticate(user=self.users["user_a"])
        response = client.get("/a/?format=json", follow=True)

        self.assertEqual(response.data, [
            {
                "name": "Test1",
                "id":1,
                "nested_dict": {
                    "secret" : {
                        "hidden" : "data"
                    },
                    "something": "public"
                }
            }
        ])

        client = APIClient()
        client.force_authenticate(user=self.users["user_b"])
        response = client.get("/a/?format=json", follow=True)

        self.assertEqual(response.data, [
            {
                "name": "Test1",
                "id":1,
                "nested_dict": {
                    "something": "public"
                }
            }
        ])

    def test_grainy_viewset_retrieve(self):
        client = APIClient()
        client.force_authenticate(user=self.users["user_a"])
        response = client.get("/a/1/?format=json", follow=True)

        self.assertEqual(response.data,
            {
                "name": "Test1",
                "id":1,
                "nested_dict": {
                    "secret" : {
                        "hidden" : "data"
                    },
                    "something": "public"
                }
            }
        )

        client = APIClient()
        client.force_authenticate(user=self.users["user_b"])
        response = client.get("/a/1/?format=json", follow=True)

        self.assertEqual(response.data,
            {
                "name": "Test1",
                "id":1,
                "nested_dict": {
                    "something": "public"
                }
            }
        )




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.shortcuts import render
from django.http import HttpResponse, JsonResponse
from django.views import View as BaseView

from rest_framework import viewsets
from django_grainy.decorators import (
    grainy_view,
    grainy_json_view,
    grainy_rest_viewset,
    grainy_rest_viewset_response,
    grainy_view_response
)

from .models import ModelA
from .serializers import ModelASerializer

# Create your views here.

@grainy_rest_viewset(
    namespace = "api.a",
    handlers = {
        "nested_dict.secret" : { "explicit" : True }
    }
)
class ModelAViewSet(viewsets.ModelViewSet):
    queryset = ModelA.objects.all()
    serializer_class = ModelASerializer


class ExplicitViewSet(viewsets.ModelViewSet):
    queryset = ModelA.objects.all()
    serializer_class = ModelASerializer

    @grainy_rest_viewset_response(
        namespace="api.a_x",
        namespace_instance="{namespace}.{instance.id}",
        explicit=True,
        explicit_instance=False
    )
    def retrieve(self,*args,**kwargs):
        return super(ExplicitViewSet, self).retrieve(*args, **kwargs)

    @grainy_rest_viewset_response(
        namespace="api.a_x",
        namespace_instance="{namespace}.{instance.id}",
        explicit=True
    )
    def destroy(self,*args,**kwargs):
        return super(ExplicitViewSet, self).destroy(*args, **kwargs)



@grainy_view(namespace="site.view")
def view(request):
    return HttpResponse()


@grainy_view(namespace="detail.{id}")
def detail(request, id):
    return HttpResponse("ID {}".format(id))


@grainy_view(
    namespace="detail.{id}",
    explicit=True,
    ignore_grant_all=True
)
def detail_explicit(request, id):
    return HttpResponse("ID {}".format(id))


@grainy_view(namespace="site.view")
class View(BaseView):
    def get(self, request):
        return HttpResponse(content="GET Response")

    def post(self, request):
        return HttpResponse(content="POST Response")

    def put(self, request):
        return HttpResponse(content="PUT Response")

    def delete(self, request):
        return HttpResponse(content="DELETE Response")

    def patch(self, request):
        return HttpResponse(content="PATCH Response")

@grainy_view(namespace="detail.{id}")
class Detail(BaseView):

    def get(self, request, id):
        return HttpResponse(content="GET Response {}".format(id))

    def post(self, request, id):
        return HttpResponse(content="POST Response {}".format(id))

    def put(self, request, id):
        return HttpResponse(content="PUT Response {}".format(id))

    def delete(self, request, id):
        return HttpResponse(content="DELETE Response {}".format(id))

    def patch(self, request, id):
        return HttpResponse(content="PATCH Response {}".format(id))


@grainy_view(
    namespace="detail.{id}",
    explicit=True,
    ignore_grant_all=True
)
class DetailExplicit(Detail):
    pass


@grainy_view(namespace="detail_manual.{id}")
class DetailManual(BaseView):

    def get(self, request, id):
        return HttpResponse(content="GET Response {}".format(id))

    @grainy_view_response(namespace="detail_manual")
    def post(self, request):
        return HttpResponse(content="POST Response 1")

@grainy_view(namespace="detail_manual.{request.method}.{id}")
class DetailReqFmt(Detail):
    pass


@grainy_json_view(
    namespace="site.view",
    handlers = {
        "nested_dict.secret" : { "explicit" : True }
    }
)
class JsonView(BaseView):
    def get(self, request):
        return JsonResponse({
            "hello" : "world",
            "nested_dict" : {
                "public": "something",
                "secret": "hidden"
            }
        })



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import six
import inspect

from django.db import models
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.utils.translation import ugettext_lazy as _

from grainy.const import PERM_READ
from grainy.core import PermissionSet, Namespace

from .fields import PermissionField
from .conf import PERM_CHOICES
from .helpers import (
    namespace,
    int_flags
)


class PermissionQuerySet(models.QuerySet):
    """
    Queryset that offers additional utilities for UserPermission and
    GroupPermission queries
    """

    def permission_set(self):
        """
        Builds and returns a grainy.PermissionSet object from all
        the rows in the query

        Returns:
            - grainy.PermissionSet
        """
        pset = PermissionSet()
        for row in self:
            pset[row.namespace] = row.permission
        return pset

class PermissionManager(models.Manager):
    """
    Object Manager that offers additional utilities for managing
    UserPermission and GroupPermission objects
    """

    def get_queryset(self):
        return PermissionQuerySet(self.model, using=self._db)

    def add_permission_set(self, pset):
        """
        Add all permissions specified in a PermissionSet

        Arguments:
            - pset <grainy.PermissionSet|dict>: if passed as `dict` permission
                values may be passed as string flags
        """

        _pset = self.permission_set()

        if not isinstance(pset, PermissionSet) and isinstance(pset, dict):
            pset = PermissionSet({ns:int_flags(f) for ns,f in pset.items()})

        for namespace, permission in pset.permissions.items():
            _pset[namespace] = permission

        for namespace, permission in _pset.permissions.items():
            perm = self.update_or_create(
                namespace = namespace,
                defaults = { "permission" : permission.value }
            )

    def add_permission(self, target, permission):
        """
        Add permission for the specified target

        Arguments:
            - target <object|class|str>
            - permission <str|int>: permission flags
        """

        self.update_or_create(
            namespace = namespace(target),
            defaults = { "permission" : int_flags(permission) }
        )

    def delete_permission(self, target):
        """
        Remove an explicit permission rule set for the
        specified target.

        Note that this does not touch permissions granted
        by rulings higher up in the namespace path

        Arguments:
            - target <object|class|str>
        """

        self.get_queryset().filter(namespace = namespace(target)).delete()


    def permission_set(self):
        """
        Return grainy.PermissionSet instance from all rows returned
        from get_queryset()

        Returns:
            - grainy.PermissionSet
        """
        return self.get_queryset().all().permission_set()

# Create your models here.

class Permission(models.Model):
    class Meta(object):
        abstract = True

    namespace = models.CharField(
        max_length=255,
        help_text=_("Permission namespace (A '.' delimited list of keys")
    )
    permission = PermissionField(
        default=PERM_READ
    )

    def __unicode__(self):
        return u"{}: {}".format(self.namespace, self.permission)

class UserPermission(Permission):

    """
    Describes permission for a user
    """
    class Meta(object):
        verbose_name = _("User Permission")
        verbose_name_plural = _("User Permissions")
        base_manager_name = "objects"

    user = models.ForeignKey(get_user_model(), related_name="grainy_permissions", on_delete=models.CASCADE)
    objects = PermissionManager()

class GroupPermission(Permission):
    """
    Describes permission for a user group
    """
    class Meta(object):
        verbose_name = _("Group Permission")
        verbose_name_plural = _("Group Permissions")
        base_manager_name = "objects"

    group = models.ForeignKey(Group, related_name="grainy_permissions", on_delete=models.CASCADE)
    objects = PermissionManager()


class GrainyHandler(object):
    """
    The base class to use for the Grainy Meta class
    """

    parent = None
    namespace_base = None
    namespace_instance_template = u"{namespace}.{instance}"

    @classmethod
    def namespace_instance(cls, instance, **kwargs):
        """
        Returns the permissioning namespace for the passed instance

        Arguments:
            - instance <object|str|Namespace>: the value of this will be appended
                to the base namespace and returned

        Keyword Arguments:
            - any keyword arguments will be used for formatting of the
                namespace

        Returns:
            - unicode: namespace
        """

        if not isinstance(cls.namespace_base, Namespace):
            raise ValueError("`namespace_base` needs to be a Namespace instance")

        return cls.namespace_instance_template.format(
            namespace=str(cls.namespace_base).format(**kwargs),
            instance=instance,
            **kwargs
        ).lower()

    @classmethod
    def namespace(cls, instance=None, **kwargs):
        """
        Wrapper function to return either the result of namespace_base or
        namespace instance depending on whether or not a value was passed in
        `instance`

        All keyword arguments will be available while formatting the
        namespace string.

        Keyword Arguments:
            - instance <object|str|Namespace>: the value of this will be appended

        Returns:
            - unicode
        """
        if instance:
            return cls.namespace_instance(instance, **kwargs)
        namespace = u"{}".format(cls.namespace_base)
        if kwargs:
            namespace = namespace.format(**kwargs)
        return namespace.lower()

    @classmethod
    def set_namespace_base(cls, value):
        if not isinstance(value, Namespace):
            raise TypeError("`value` needs to be a Namespace instance")
        cls.namespace_base = value

    @classmethod
    def set_parent(cls, parent):
        cls.parent = parent

class GrainyModelHandler(GrainyHandler):

    """
    grainy model handler meta class
    """

    model = None
    namespace_instance_template = u"{namespace}.{instance.id}"

    @classmethod
    def set_parent(cls, model):
        cls.parent = model
        cls.model = model
        cls.set_namespace_base(
            Namespace([
                model._meta.app_label,
                model._meta.object_name
            ])
        )


class GrainyMixin(object):

    @property
    def grainy_namespace(self):
        return self.Grainy.namespace(self)



import grainy.const
from django.conf import settings
from .const import PERM_CHOICES_CRUD

DJANGO_OP_TO_FLAG = getattr(settings, "GRAINY_DJANGO_OP_TO_FLAG", {
    "add" : grainy.const.PERM_CREATE,
    "delete" : grainy.const.PERM_DELETE,
    "change" : grainy.const.PERM_UPDATE,
    "view" : grainy.const.PERM_READ
})

REQUEST_METHOD_TO_FLAG = getattr(settings, "GRAINY_REQUEST_METHOD_TO_FLAG", {
    "HEAD" : grainy.const.PERM_READ,
    "OPTIONS" : grainy.const.PERM_READ,
    "GET" : grainy.const.PERM_READ,
    "PUT" : grainy.const.PERM_UPDATE,
    "PATCH" : grainy.const.PERM_UPDATE,
    "POST" : grainy.const.PERM_CREATE,
    "DELETE" : grainy.const.PERM_DELETE,
})

ANONYMOUS_PERMS = getattr(settings, "GRAINY_ANONYMOUS_PERMS", {
})

PERM_CHOICES = getattr(settings, "GRAINY_PERM_CHOICES", PERM_CHOICES_CRUD)
PERM_CHOICES_FOR_FIELD = [(n,v) for n,v,c in PERM_CHOICES]

# if True, the default django forms for adding / editing permissions
# will be removed.
ADMIN_REMOVE_DEFAULT_FORMS = getattr(
    settings, "GRAINY_ADMIN_REMOVE_DEFAULT_FORMS", True
)



import re
import logging


from django.contrib.auth.backends import ModelBackend
from django.contrib.auth import get_user_model

from .helpers import (
    django_op_to_flag,
)

from .models import (
    namespace,
)

from .util import (
    Permissions,
)


class GrainyBackend(ModelBackend):

    """
    Authenticate actions using grainy permissions
    """

    def has_module_perms(self, user, obj=None):

        # superusers have access to everything
        if user.is_superuser:
            return True

        return Permissions(user).check(obj, django_op_to_flag("view"))

    def has_perm(self, user, perm, obj=None):

        # superusers have access to everything
        if user.is_superuser:
            return True

        ns = None
        if obj:
            try:
                ns = namespace(obj)
            except TypeError as inst:
                ns = None

        try:
            label, action = tuple(perm.split("."))
            a = re.match("(add|delete|change|view)_(.+)", action)
        except ValueError as inst:
            a = None

        if a:
            flag = django_op_to_flag(a.group(1))
            if not ns:
                ns = "{}.{}".format(label, a.group(2))
        else:
            flag = django_op_to_flag("view")
            if not ns:
                ns = perm

        return Permissions(user).check(ns, flag)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.contrib import admin

from .models import ModelA, ModelB

# Register your models here.

@admin.register(ModelA)
class ModelAAdmin(admin.ModelAdmin):
    list_display = ("id", "name")



import six
from django.db import models
from django import forms
from .conf import PERM_CHOICES

class PermissionFormField(forms.IntegerField):
    def prepare_value(self, value):
        # if the form field is passed a bitmask we
        # need to convert it to a list, where each
        # item represents a choice (flag)
        if isinstance(value, six.integer_types):
            _value = []
            for f,n,c in PERM_CHOICES:
                if value & f:
                    _value.append(f)
            value = _value
        else:
            value = value or 0
        return value

    def clean(self, value):
        if isinstance(value, list):
            _value = 0
            for flag in value:
                _value |= int(flag)
            value = _value
        return value



class PermissionField(models.IntegerField):

    def to_python(self, value):
        # if a string is passed it should be parsed
        # for string flags (for example 'c', 'r', 'u' and 'd')
        # as specified in the PERM_CHOICES setting and
        # convert to bitmask
        if isinstance(value, six.string_types):
            _value = 0
            for flag, name, str_flag in PERM_CHOICES:
                if str_flag in value:
                    _value |= flag
            value = _value
        elif value is None:
            return 0

        return value




from .util import UserTestCase
from django.db import models
from django_grainy.models import UserPermission, GroupPermission, GrainyHandler
from django_grainy.decorators import grainy_model

from django_grainy_test.models import (
    ModelA,
    ModelB,
    ModelC,
    ModelD,
    ModelX,
    ModelY,
    ModelZ
)

from grainy.const import (
    PERM_READ,
    PERM_UPDATE,
    PERM_CREATE,
    PERM_DELETE
)


class TestGrainyHandler(UserTestCase):

    @classmethod
    def setUpTestData(cls):
        UserTestCase.setUpTestData()

    def test_grainy_model_decorator(self):
        a = ModelA.objects.create(name="A1")
        self.assertEqual(a.Grainy.namespace(a), "django_grainy_test.modela.{}".format(a.id))
        b = ModelB.objects.create(name="B1")
        self.assertEqual(b.Grainy.namespace(b), "something.arbitrary.{}".format(b.id))

        c = ModelC.objects.create(name="C1",b=b)
        self.assertEqual(c.Grainy.namespace(), ModelB.Grainy.namespace())
        self.assertEqual(c.Grainy.namespace(c), "something.arbitrary.2.c.3")


        d = ModelD.objects.create(name="D1")
        self.assertEqual(d.Grainy.namespace(value="parent"), "dynamic.parent")
        self.assertEqual(
            d.Grainy.namespace(d, value="parent", other_value="child"),
            "dynamic.parent.child"
        )

        x = ModelX.objects.create(name="X1")
        self.assertEqual(x.Grainy.namespace(), "x")
        self.assertEqual(x.Grainy.namespace(x), "x.{}".format(x.id))

        y = ModelY.objects.create(name="Y1",x=x)
        self.assertEqual(y.Grainy.namespace(), "custom")
        self.assertEqual(y.Grainy.namespace_instance_template, "x.{instance.x.pk}.custom.{instance.pk}")
        self.assertEqual(y.Grainy.namespace(y), "x.5.custom.6")

        z = ModelZ.objects.create(name="Z1",y=y)
        self.assertEqual(z.Grainy.namespace_instance_template, "x.{instance.y.x.pk}.custom.{instance.y.pk}.z.{instance.pk}")
        self.assertEqual(z.Grainy.namespace(z), "x.5.custom.6.z.7")


    def test_grainy_mixin(self):
        a = ModelA.objects.create(name="A1")
        self.assertEqual(a.grainy_namespace, "django_grainy_test.modela.{}".format(a.id))


